from ..test_helper import TestConfig
from .tableau import TableauServerResultTarget
from .tableau import TableauOnlineResultTarget

from unittest import TestCase
from nose.tools import eq_, raises

from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import parse_qs
from six.moves.urllib.parse import unquote as url_unquote

test_config = TestConfig()

class TestTableauServerResultTarget(TableauServerResultTarget):
    server = 'tableau.example.com'
    username = 'test@example.com'
    password = 'test-password'

class TableauServerResultTargetTestCase(TestCase):
    def setUp(self):
        test_config.setUp()

    def tearDown(self):
        test_config.tearDown()

    def test_default(self):
        target = TestTableauServerResultTarget(test_config.get_tmp_path('result.job'))
        target.datasource = 'test-datasource'
        url = urlparse(target.get_result_url())
        params = parse_qs(url.query)
        eq_(url.scheme, 'tableau')
        eq_(url.hostname, 'tableau.example.com')
        eq_(url.path, '/' + target.datasource)
        eq_(url_unquote(url.username), TestTableauServerResultTarget.username)
        eq_(url_unquote(url.password), TestTableauServerResultTarget.password)
        eq_(params.get('ssl'), ['true'])
        eq_(params.get('ssl_verify'), ['true'])
        eq_(params.get('server_version'), None)
        eq_(params.get('site'), None)
        eq_(params.get('project'), None)
        eq_(params.get('mode'), ['replace'])

    def test_options(self):
        target = TestTableauServerResultTarget(test_config.get_tmp_path('result.job'))
        target.server_version = '8.3'
        target.site = 'test-site'
        target.project = 'test-project'
        target.datasource = 'test-datasource'
        target.mode = 'append'
        url = urlparse(target.get_result_url())
        params = parse_qs(url.query)
        eq_(params.get('server_version'), [target.server_version])
        eq_(params.get('site'), [target.site])
        eq_(params.get('project'), [target.project])
        eq_(params.get('mode'), [target.mode])

class TableauOnlineResultTargetTestCase(TestCase):
    def setUp(self):
        test_config.setUp()

    def tearDown(self):
        test_config.tearDown()

    def test_default(self):
        target = TableauOnlineResultTarget(test_config.get_tmp_path('result.job'))
        target.username = 'test@example.com'
        target.password = 'test-password'
        target.datasource = 'test-datasource'
        url = urlparse(target.get_result_url())
        params = parse_qs(url.query)
        eq_(url.scheme, 'tableau')
        eq_(url.hostname, 'online.tableausoftware.com')
        eq_(url.path, '/' + target.datasource)
        eq_(url_unquote(url.username), target.username)
        eq_(url_unquote(url.password), target.password)
        eq_(params['server_version'], ['online'])
        eq_(params['mode'], ['replace'])



from luigi_td.bulk_import import BulkImport
from luigi_td.client import ResultProxy
from luigi_td.config import Config, get_config
from luigi_td.task import DatabaseTask, TableTask, Query
from luigi_td.targets.result import ResultTarget
from luigi_td.targets.s3 import S3ResultTarget
from luigi_td.targets.tableau import TableauServerResultTarget, TableauOnlineResultTarget
from luigi_td.targets.td import DatabaseTarget, TableTarget, SchemaError

__all__ = [
    # bulk_import
    'BulkImport',
    # client
    'ResultProxy',
    # config
    'Config',
    'get_config',
    # task
    'DatabaseTask',
    'TableTask',
    'Query',
    # targets.result
    'ResultTarget',
    # targets.s3
    'S3ResultTarget',
    # targets.tableau
    'TableauServerResultTarget',
    'TableauOnlineResultTarget',
    # targets.td
    'DatabaseTarget',
    'TableTarget',
    'SchemaError',
]



import os
import luigi
import tdclient

import logging
logger = logging.getLogger('luigi-interface')

DEFAULT_ENDPOINT = 'https://api.treasuredata.com/'

class Config(object):
    def __init__(self, apikey, endpoint=DEFAULT_ENDPOINT):
        self.apikey = apikey
        self.endpoint = endpoint
        if not self.endpoint.endswith('/'):
            self.endpoint += '/'

    def get_client(self):
        return tdclient.Client(self.apikey, endpoint=self.endpoint)

class ConfigLoader(object):
    _instance = None

    @classmethod
    def instance(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
            cls._instance.load_default()
        return cls._instance

    def __init__(self):
        self.config = None

    def get_config(self):
        return self.config

    def load_default(self):
        luigi_config = luigi.configuration.get_config()
        apikey = luigi_config.get('td', 'apikey', os.environ.get('TD_API_KEY'))
        endpoint = luigi_config.get('td', 'endpoint', DEFAULT_ENDPOINT)
        self.config = Config(apikey, endpoint=endpoint)

def get_config():
    return ConfigLoader.instance().get_config()



from luigi_td.config import get_config
from luigi_td.client import ResultProxy
from luigi_td.targets.result import ResultTarget
from luigi_td.targets.td import DatabaseTarget
from luigi_td.targets.td import TableTarget

import time
import luigi
import jinja2

import logging
logger = logging.getLogger('luigi-interface')

class DatabaseTask(luigi.Task):
    config = get_config()
    database_name = luigi.Parameter()
    action = luigi.Parameter(default='create')

    def output(self):
        return DatabaseTarget(self.database_name)

    def run(self):
        client = self.config.get_client()
        logger.debug('%s: creating database: %s', self, self.database_name)
        client.create_database(self.database_name)

class TableTask(luigi.Task):
    config = get_config()
    database_name = luigi.Parameter()
    table_name = luigi.Parameter()
    action = luigi.Parameter(default='create')
    schema = luigi.Parameter(default=None, significant=False)
    empty = luigi.BoolParameter(default=False, significant=False)

    def requires(self):
        return DatabaseTask(self.database_name)

    def output(self):
        return TableTarget(self.database_name, self.table_name, self.schema, empty=self.empty)

    def run(self):
        client = self.config.get_client()
        logger.debug('%s: creating table: %s.%s', self, self.database_name, self.table_name)
        client.create_log_table(self.database_name, self.table_name)
        if self.schema is not None:
            logger.debug('%s: updating schema for %s.%s', self, self.database_name, self.table_name)
            client.update_schema(self.database_name, self.table_name, [s.split(':') for s in self.schema])

# query

class Timeout(Exception):
    pass

class Query(luigi.Task):
    config = get_config()
    debug = False
    timeout = None
    priority = None
    retry_limit = None
    type = 'hive'
    database = None
    source = None
    variables = {}

    def query(self):
        return NotImplemented

    def load_query(self, source):
        env = jinja2.Environment(loader=jinja2.PackageLoader(self.__module__, '.'))
        template = env.get_template(source)
        return template.render(task=self, **self.variables)

    def run_query(self, query):
        result = self.output()
        result_url = None
        if isinstance(result, ResultTarget):
            result_url = result.get_result_url()
        client = self.config.get_client()
        job = client.query(self.database,
                           query,
                           priority = self.priority,
                           retry_limit = self.retry_limit,
                           type = self.type,
                           result_url = result_url)
        job.update()
        logger.info("%s: td.job.url: %s", self, job.url)

        # wait for the result
        try:
            if self.timeout:
                timeout = time.time() + self.timeout
            else:
                timeout = None
            while not job.finished():
                if timeout and time.time() > timeout:
                    raise Timeout('{0} (Job ID {1}) timed out after {2} seconds'.format(
                        self, job.job_id, self.timeout))
                time.sleep(2)
        except:
            # kill query on exceptions
            job.kill()
            raise
        job.update()

        logger.info("%s: td.job.result: job_id=%s status=%s", self, job.job_id, job.status())

        if not job.success():
            stderr = job.debug['stderr']
            if stderr:
                logger.error(stderr)
            raise RuntimeError("job {0} {1}\n\nOutput:\n{2}".format(job.job_id, job.status(), job.debug['cmdout']))

        return ResultProxy(job)

    def run(self):
        if hasattr(self, 'query_file'):
            self.source = self.query_file
        query = self.load_query(self.source) if self.source else self.query()
        result = self.run_query(query)
        target = self.output()
        if target and isinstance(target, ResultTarget):
            target.save_result_state(result)



from .bulk_import import BulkImportUploadContext
from .bulk_import import BulkImportTarget
from .bulk_import import BulkImportSession
from .bulk_import import BulkImportUpload
from .bulk_import import BulkImportPerform
from .bulk_import import BulkImport
from .test_helper import MockJob
from .test_helper import TestConfig

from unittest import TestCase
from nose.tools import ok_, eq_, raises

import json
import os
import shutil
import tempfile
import luigi

test_config = TestConfig(
    jobs = [
        {
            'job_id': 1,
            'status': 'success',
        }
    ],
    bulk_imports = [
        {
            'session': 'session-1',
            'status': 'uploading',
            'frozen': False,
        },
        {
            'session': 'session-2',
            'status': 'uploading',
            'frozen': True,
        },
        {
            'session': 'session-3',
            'status': 'performing',
            'frozen': True,
            'job_id': 1,
        },
        {
            'session': 'session-4',
            'status': 'ready',
            'frozen': True,
            'job_id': 1,
        },
        {
            'session': 'session-5',
            'status': 'committing',
            'frozen': True,
            'job_id': 1,
        },
        {
            'session': 'session-6',
            'status': 'committed',
            'frozen': True,
            'job_id': 1,
        },
    ]
)

class BulkImportUploadContextTestCase(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_creates_temporary_environ(self):
        env = {'TEST_ENV': 'ok'}
        eq_(os.environ.get('TEST_ENV'), None)
        with BulkImportUploadContext(env):
            eq_(os.environ.get('TEST_ENV'), 'ok')
        eq_(os.environ.get('TEST_ENV'), None)

    def test_creates_directories(self):
        cur_dir = os.getcwd()
        log_dir = os.path.realpath(os.path.join(self.test_dir, 'log'))
        tmp_dir = os.path.realpath(os.path.join(self.test_dir, 'tmp'))
        with BulkImportUploadContext(log_dir=log_dir, tmp_dir=tmp_dir):
            eq_(os.getcwd(), log_dir)
            eq_(os.path.exists(log_dir), True)
            eq_(os.path.exists(tmp_dir), True)
        eq_(os.getcwd(), cur_dir)
        eq_(os.path.exists(log_dir), True)
        eq_(os.path.exists(tmp_dir), False)

class TestBulkImportTarget(BulkImportTarget):
    config = test_config

class BulkImportTargetTestCase(TestCase):
    def test_not_exists(self):
        target = TestBulkImportTarget('none')
        eq_(target.exists(), False)

    def test_exists(self):
        target = TestBulkImportTarget('session-1')
        eq_(target.exists(), True)

class TestBulkImportSession(BulkImportSession):
    config = test_config

class BulkImportSessionTestCase(TestCase):
    test_options = {
        'database': 'test_db',
        'table': 'test_tb',
    }

    def test_output(self):
        task = TestBulkImportSession('session-1', options=json.dumps(self.test_options))
        eq_(type(task.output()), BulkImportTarget)

    def test_run(self):
        task = TestBulkImportSession('session-1', options=json.dumps(self.test_options))
        task.run()

class TestBulkImportUpload(BulkImportUpload):
    config = test_config

class BulkImportUploadTestCase(TestCase):
    test_options = {
        'database': 'test_db',
        'table': 'test_tb',
        'path': 'test_input.tsv',
        'format': 'tsv',
        'columns': ['date:string', 'x:int', 'y:int'],
        'time_column': 'date',
        'time_format': '%Y-%m-%d',
    }

    def test_requires(self):
        task = TestBulkImportUpload('session-1', options=json.dumps(self.test_options))
        eq_(type(task.requires()), BulkImportSession)

    def test_complete(self):
        task = TestBulkImportUpload('none', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportUpload('session-1', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportUpload('session-2', options=json.dumps(self.test_options))
        eq_(task.complete(), True)

    def test_command_line(self):
        task = TestBulkImportUpload('session-1', options=json.dumps(self.test_options))
        env, args = task.build_command_line()
        # env
        eq_(env['TD_TOOLBELT_JAR_UPDATE'], '0')
        # args
        ok_(task.config.endpoint in args)
        ok_(task.config.apikey in args)
        ok_('import:upload' in args)
        ok_('session-1' in args)
        ok_(os.path.join(task.tmp_dir, 'parts') in args)
        ok_(os.path.join(task.log_dir, 'error-records') in args)
        ok_(self.test_options['path'] in args)
        ok_(self.test_options['format'] in args)
        ok_(self.test_options['time_column'] in args)
        ok_(self.test_options['time_format'] in args)
        ok_('date,x,y' in args)
        ok_('string,int,int' in args)
        ok_('--empty-as-null-if-numeric' in args)

class TestBulkImportPerform(BulkImportPerform):
    config = test_config

class BulkImportPerformTestCase(TestCase):
    test_options = {
        'database': 'test_db',
        'table': 'test_tb',
    }

    def test_requires(self):
        task = TestBulkImportPerform('session-1', options=json.dumps(self.test_options))
        eq_(type(task.requires()), BulkImportUpload)

    def test_complete(self):
        task = TestBulkImportPerform('none', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportPerform('session-1', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportPerform('session-2', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportPerform('session-3', options=json.dumps(self.test_options))
        eq_(task.complete(), False)
        task = TestBulkImportPerform('session-4', options=json.dumps(self.test_options))
        eq_(task.complete(), True)

    def test_run_2(self):
        task = TestBulkImportPerform('session-2', options=json.dumps(self.test_options))
        task.run()

    def test_run_3(self):
        task = TestBulkImportPerform('session-3', options=json.dumps(self.test_options))
        task.run()

    def test_run_4(self):
        task = TestBulkImportPerform('session-4', options=json.dumps(self.test_options))
        task.run()

class TestBulkImportInput(luigi.ExternalTask):
    def output(self):
        return luigi.LocalTarget('test_input.tsv')

class TestBulkImport(BulkImport):
    config = test_config
    database = 'test_db'
    table = 'test_tb'
    format = 'tsv'
    columns = ['date:string', 'x:int', 'y:int']
    time_column = 'date'
    time_format = '%Y-%m-%d'

    def requires(self):
        return TestBulkImportInput()

class BulkImportTestCase(TestCase):
    def test_complete(self):
        task = TestBulkImport('none')
        eq_(task.complete(), False)
        task = TestBulkImport('session-1')
        eq_(task.complete(), False)
        task = TestBulkImport('session-2')
        eq_(task.complete(), False)
        task = TestBulkImport('session-3')
        eq_(task.complete(), False)
        task = TestBulkImport('session-4')
        eq_(task.complete(), False)
        task = TestBulkImport('session-5')
        eq_(task.complete(), False)
        task = TestBulkImport('session-6')
        eq_(task.complete(), True)

    def test_local_path(self):
        class TestInput(luigi.ExternalTask):
            def output(self):
                return luigi.LocalTarget('test_input.tsv')
        class TestBulkImportFromLocalTarget(TestBulkImport):
            def requires(self):
                return TestInput()
        task = TestBulkImportFromLocalTarget('session-1')
        eq_(task.get_path(), os.path.realpath('test_input.tsv'))

    def test_s3_path(self):
        class TestInput(luigi.ExternalTask):
            def output(self):
                s3client = luigi.s3.S3Client(aws_access_key_id='test-key', aws_secret_access_key='test-secret')
                return luigi.s3.S3Target('s3://test-bucket/test_input.tsv', client=s3client)
        class TestBulkImportFromS3Target(TestBulkImport):
            def requires(self):
                return TestInput()
        task = TestBulkImportFromS3Target('session-1')
        eq_(task.get_path(), 's3://test-key:test-secret@/test-bucket/test_input.tsv')

    def test_run_without_commit(self):
        task = TestBulkImport('session-4', no_commit=True)
        for req in task.run():
            eq_(type(req), BulkImportPerform)

    def test_run(self):
        task = TestBulkImport('session-4')
        for req in task.run():
            eq_(type(req), BulkImportPerform)



from .config import Config, ConfigLoader, get_config, DEFAULT_ENDPOINT

from unittest import TestCase
from nose.tools import eq_, raises

import os
import luigi
import tdclient

class ConfigTestCase(TestCase):
    def test_create(self):
        config = Config('test-apikey')
        eq_(type(config.get_client()), tdclient.Client)
        eq_(config.apikey, 'test-apikey')
        eq_(config.endpoint, DEFAULT_ENDPOINT)

    def test_create_with_endpoint(self):
        config = Config('test-apikey', endpoint='https://api.example.com')
        eq_(type(config.get_client()), tdclient.Client)
        eq_(config.apikey, 'test-apikey')
        eq_(config.endpoint, 'https://api.example.com/')

class ConfigLoaderTestCase(TestCase):
    def setUp(self):
        self.environ = os.environ.copy()
        # clear environment variables
        if 'TD_API_KEY' in os.environ:
            del os.environ['TD_API_KEY']

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self.environ)
        if os.path.exists('client.cfg'):
            os.unlink('client.cfg')
            luigi.configuration.LuigiConfigParser._instance = None

    def _config(self, values):
        with open('client.cfg', 'w') as f:
            f.write("[td]\n")
            for key, val in values.items():
                f.write("{0}: {1}\n".format(key, val))
        luigi.configuration.LuigiConfigParser._instance = None

    def _get_config(self):
        loader = ConfigLoader()
        loader.load_default()
        return loader.get_config()

    @raises(ValueError)
    def test_no_config(self):
        config = self._get_config()
        eq_(config.apikey, None)
        config.get_client()

    def test_apikey_by_environ(self):
        os.environ['TD_API_KEY'] = 'test-apikey'
        config = self._get_config()
        eq_(config.apikey, 'test-apikey')

    def test_apikey_by_luigi_config(self):
        self._config({'apikey': 'test-apikey'})
        config = self._get_config()
        eq_(config.apikey, 'test-apikey')

class GetConfigTestCase(TestCase):
    def test_default(self):
        eq_(type(get_config()), Config)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="luigi-td",
    version='0.6.9',
    description="Luigi integration for Treasure Data",
    author="Treasure Data, Inc.",
    author_email="support@treasure-data.com",
    url="https://github.com/treasure-data/luigi-td",
    install_requires=open("requirements.txt").read().splitlines(),
    packages=find_packages(),
    license="Apache License 2.0",
    platforms="Posix; MacOS X; Windows",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
)
